#!/usr/bin/env python

import os
import sys
import warnings
import pandas as pd 
import numpy as np 
import argparse # pour gerer les options
import re # pour les regexp
from pprint import pprint


class gmapstats (object):

	def __init__(self):
		# initialize mains objects here
		self.parameters = None
		self.gmapResults = {}
		self.perfectMatches = []
		self.ambigousMatches = []

		self.stats = { 'num_perfect_matches' : 0, 
			'num_perfect_multi' : 0,
			'num_ambigous_match':0,
			'num_ambigous_multi':0}

		self.version='1.0'
		self.author='Helene Rimbert'

	def main(self):
		"""
		MAIN
		"""
		self.parseOptions()
		self.createWorkingEnv()
		self.getGmapResults()
		self.parseGmapResults()

		pprint(self.stats)

		self.outputPerfectMatch.close()
		self.outputPerfectMatchMulti.close()
		self.outputAmbigousMatch.close()
		self.outputAmbigousMatchMulti.close()


	def parseOptions(self):
		"""
		Use Argparse to setup/get options from cmd line
		"""
		parser = argparse.ArgumentParser()
		parser.add_argument("-i", "--input", help="GMAP results in PSL format (-f 1 gmap option)", required=True)
		parser.add_argument("-o", "--outputdir", help="output directory", required=True)
		parser.add_argument("-c", "--qcov", help="min query coverage for ambigous matches", default=0.95)
		self.parameters=parser.parse_args()

	def createWorkingEnv(self):

		# create output dir
		try:
			os.mkdir(self.parameters.outputdir)
		except OSError:
			sys.stderr.write(" Creation of directory  %s failed" % self.parameters.outputdir)
		else:
			sys.stderr.write(" Successfully created the directory %s " % self.parameters.outputdir)

		# create filehandler for results
		self.outputPerfectMatch = open(self.parameters.outputdir+"/perfectmatch.psl", "w")
		self.outputPerfectMatchMulti = open(self.parameters.outputdir+"/perfectmatch_multi.psl", "w")
		self.outputAmbigousMatch = open(self.parameters.outputdir+"/ambigous_match.psl", "w")
		self.outputAmbigousMatchMulti = open(self.parameters.outputdir+"/ambigous_match_multi.psl", "w")
		self.outputBelowThreshold = open(self.parameters.outputdir+"/belowThreshold.psl", "w")


	def getGmapResults(self):
		"""
		Read gmap psl results file and save into dict
		"""

		pslFile = self.parameters.input
		if os.path.isfile(pslFile):
			with open(pslFile, 'r') as myFile:
				pslContent=myFile.readlines()

			for pslLine in pslContent:
				(matches,mismatches,repmatches,ncount,qnuminsert,qbaseinsert,tnuminsert,tbaseinsert,strand,qname,qsize,qstart,qend,tname,tsize,tstart,tend,blockcount,blocksize,qstart,tstart) = pslLine.split("\t")
				
				if qname in self.gmapResults:
					#print("     + new gmap results for query %s " % str(qname))
					self.gmapResults[qname].append(pslLine)
				else:
					#print(" >>>>>>>>>> new Query %s " % str(qname))
					self.gmapResults[qname] = [pslLine]


	def getInitialChrom(self, query):
		"""
		extract chromosome from query name
		"""
		queryinfo = query.split(':')
		chrom=queryinfo[2]
		return chrom

	def parseGmapResults(self):
		"""
		foreach query, parse result and check if we are having a perfect match or not
		"""

		queryList = list(self.gmapResults.keys())
		numQueries = len(queryList)
		isHit = True
		sys.stderr.write(" Found %i queries in GMAP results " % numQueries)

		for query in queryList:
			if query == "qname":
				continue

			#print(' \n\n_____________________________________________________________')
			#print(' |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||')
			#print(' _____________________________________________________________')
			#print( " Evaluate query %s" % str(query))

			# get dict of gmap results
			gmapresults = self.gmapResults[query]
			nhits = len(gmapresults)
			#print(" %i results found" % nhits)
			#pprint(gmapresults)

			# create dict of perfect hit(s) and ambigous hits for current query
			self.perfectMatches = []
			self.ambigousMatches = []

			# loop over all results
			for hitIndex in range(len(gmapresults)):
				hit=gmapresults[hitIndex]
				#print(" Hit index: %s" % str(hitIndex))
				(matches,mismatches,repmatches,ncount,qnuminsert,qbaseinsert,tnuminsert,tbaseinsert,strand,qname,qsize,qstart,qend,tname,tsize,tstart,tend,blockcount,blocksize,qstart,tstart) = hit.split("\t")
				
				# get percentage of query coverage
				qcoverage = (int(matches) / int(qsize)) * 100 

				if qcoverage < self.parameters.qcov :
					#print(" Query coverage lower than threshold (%s): %s" % str(self.parameters.qcov), str(qcoverage))
					#print(" Cannot anchor gene %s" % str(query))
					self.outputMatches(matches=hit, category='belowThreshold')
					isHit = False
					break
				else:
					#print(" Hit above coverage for query %s : continue" % str(query))
					isHit = True

				# extract chromosome from query (based on query name)
				initialChrom = self.getInitialChrom(query=qname)

				# check if current hit is a perfect match
				isPerfect = self.isPerfectMatch(hit=hit, qcov=qcoverage, initChrom=initialChrom)

				if isPerfect is True:
					self.perfectMatches.append(hit)

				else:

					if  str(hitIndex) == "0" :
						if isPerfect is False:
							#print(" First hit is not perfect:")
							#pprint(hit)
							self.ambigousMatches.append(hit)
						else:
							sys.stderr.write(" WARNING: 1st hit perfect uncorrectly managed !!")

					else:

						if (len(self.perfectMatches)>0):
							#print(" Already found a perfect Hit for query %s" % str(query))
							#pprint(self.perfectMatches)
							#print(" going to next query")
							break
						
						else:
							#print(" case where no perfect match and not first hit")
							isIdentical = self.isIdenticalToPreviousHit(hit=hit, hitRef=self.ambigousMatches[-1])

							if isIdentical is True:
								self.ambigousMatches.append(hit)
							else:
								break

			
			if isHit is True:
				numPerfectMatches = len(self.perfectMatches)
				sys.stderr.write(" found %i perfect matches" % numPerfectMatches)
				#pprint(self.perfectMatches)
				numAmbigousMatches = len(self.ambigousMatches)
				sys.stderr.write(" found %i ambigous matches" % numAmbigousMatches)
				#pprint(self.ambigousMatches)


				if numPerfectMatches == 1:
					self.stats['num_perfect_matches'] += 1
					self.outputMatches(matches=self.perfectMatches, category='perfect')

				elif numPerfectMatches >1:
					self.stats['num_perfect_multi'] += 1
					self.outputMatches(matches=self.perfectMatches, category='multi')

				elif numAmbigousMatches == 1:
					self.stats['num_ambigous_match'] += 1
					self.outputMatches(matches=self.ambigousMatches, category='ambigous')

				elif numAmbigousMatches >= 1:
					self.stats['num_ambigous_multi'] += 1
					self.outputMatches(matches=self.ambigousMatches, category='ambigous_multi')
				else:
					sys.stderr.write(" Problem with count of ambigous and perfect matches for query %s" % str(query))

			

	def outputMatches(self,matches,category):

		if category == 'multi':
			for hit in matches:
				self.outputPerfectMatchMulti.write(str(hit))

		elif category == 'perfect':
			self.outputPerfectMatch.write(str(self.perfectMatches[0]))

		elif category == 'ambigous':
			for hit in matches:
				self.outputAmbigousMatch.write(str(hit))

		elif category == 'ambigous_multi':
			for hit in matches:
				self.outputAmbigousMatchMulti.write(hit)

		elif category == 'belowThreshold':
			self.outputBelowThreshold.write(matches)

		else:
			sys.stderr.write(" ERROR: problem with category of gmap hits")
			sys.stderr.write("        Please check ")
			#pprint(matches)
			sys.exit()

	def getInitialChrom(self, query):
		"""
		extract chromosome from query name
		"""
		queryinfo = query.split(':')
		chrom=queryinfo[2]
		return chrom.lower()	

	def isPerfectMatch(self, hit, qcov, initChrom):
		(matches,mismatches,repmatches,ncount,qnuminsert,qbaseinsert,tnuminsert,tbaseinsert,strand,qname,qsize,qstart,qend,tname,tsize,tstart,tend,blockcount,blocksize,qstart,tstart) = hit.split("\t")
		tname = tname.lower()

		# print( " -----------------------------------------")
		# print( " target: %s" % tname)
		# print( " query coverage: %s" % str(qcov))
		# print( " query initial chromosome: %s" % str(initChrom))
		# print( " target length: %s" % tsize)
		# print( " query length: %s" % qsize)
		# print( " num matches: %s" % matches)
		# print( " num mismatches: %s" % mismatches)
		# print( " num repmatches: %s" % repmatches)
		# print( " num ncount: %s" % ncount)
		# print( " num query insert: %s" % qnuminsert)
		# print( " num query base insert: %s" % qbaseinsert)
		# print( " num target insert: %s" % tnuminsert)
		# print( " num target base insert: %s" % tbaseinsert)
		# print( " num align. block: %s" % blockcount)

		if 	(int(qcov) == 100 and int(mismatches) == 0 and  int(repmatches) == 0 and  int(ncount) == 0 and  int(qnuminsert) == 0 and  int(qbaseinsert) == 0 and int(tnuminsert) == 0 and  int(tbaseinsert) == 0 and blockcount == "1") :
			"""
			perfect Match found
			"""
			sys.stderr.write("     >>>>>>>>>>>>>>>> Perfect match against target %s <<<<<<<<<<<<<<<<<<" % tname)
			return True
		else:
			#print(" match is not perfect: %s" % str(hit))
			return False

	def isIdenticalToPreviousHit(self, hitRef, hit):
		#(matches,mismatches,repmatches,ncount,qnuminsert,qbaseinsert,tnuminsert,tbaseinsert,strand,qname,qsize,qstart,qend,tname,tsize,tstart,tend,blockcount,blocksize,qstart,tstart) = hit.split("\t")
		hit_list=hit.split("\t")
		hitRef_list=hitRef.split("\t")

		# print(" comparing tw lits:")
		# print(" ref list: %s" % str(hitRef_list))
		# print(" current list: %s" % str(hit_list))

		# print(" parts of array to compare:")
		# pprint(hitRef_list[0:7])
		# print("          vs ")
		# pprint(hit_list[0:7])

		# print(" ---------- and -----------")
		# pprint(hitRef_list[17:19])
		# print("          vs ")
		# pprint(hit_list[17:19])

		if (hitRef_list[0:7] == hit_list[0:7] and hitRef_list[17:19] == hit_list[17:19]):
			#print(" Same array content: the 2 hits are the same: need to increment ambigousMatches array")
			#input()
			return True
		else:
			#print(" Arrays are not the same: second hit is worse than the previous")
			#input()
			return False


if __name__ == "__main__":
	"""
	This code block will only be executed if gmapstats.py is called as a script
	(ie. it will not be executed if the gmapstats class is just imported by another module)
	"""
	# build the script object
	engine=gmapstats()

	# run execute the main method
	engine.main()
